package com.coherentlogic.wb.client.core.builders;

import static com.coherentlogic.wb.client.core.domain.Constants.COUNTRIES;
import static com.coherentlogic.wb.client.core.domain.Constants.DATA_CATALOG;
import static com.coherentlogic.wb.client.core.domain.Constants.FREQUENCY;
import static com.coherentlogic.wb.client.core.domain.Constants.GAP_FILL;
import static com.coherentlogic.wb.client.core.domain.Constants.INCOME_LEVEL;
import static com.coherentlogic.wb.client.core.domain.Constants.INCOME_LEVELS;
import static com.coherentlogic.wb.client.core.domain.Constants.INDICATOR;
import static com.coherentlogic.wb.client.core.domain.Constants.INDICATORS;
import static com.coherentlogic.wb.client.core.domain.Constants.LENDING_TYPE;
import static com.coherentlogic.wb.client.core.domain.Constants.LENDING_TYPES;
import static com.coherentlogic.wb.client.core.domain.Constants.METATYPES;
import static com.coherentlogic.wb.client.core.domain.Constants.MRV;
import static com.coherentlogic.wb.client.core.domain.Constants.NO;
import static com.coherentlogic.wb.client.core.domain.Constants.PAGE;
import static com.coherentlogic.wb.client.core.domain.Constants.PER_PAGE;
import static com.coherentlogic.wb.client.core.domain.Constants.REGION;
import static com.coherentlogic.wb.client.core.domain.Constants.SOURCE;
import static com.coherentlogic.wb.client.core.domain.Constants.SOURCES;
import static com.coherentlogic.wb.client.core.domain.Constants.TOPIC;
import static com.coherentlogic.wb.client.core.domain.Constants.TOPICS;
import static com.coherentlogic.wb.client.core.domain.Constants.YES;

import java.util.regex.Matcher;
import java.util.regex.Pattern;

import javax.ws.rs.core.UriBuilder;

import org.apache.commons.beanutils.ConvertUtils;
import org.apache.commons.lang.builder.ToStringBuilder;
import org.springframework.web.client.RestTemplate;

import com.coherentlogic.coherent.data.model.core.builders.AbstractQueryBuilder;
import com.coherentlogic.coherent.data.model.core.util.Constants;
import com.coherentlogic.wb.client.core.domain.Frequency;
import com.coherentlogic.wb.client.core.domain.IncomeLevelCodes;
import com.coherentlogic.wb.client.core.domain.Iso2Codes;
import com.coherentlogic.wb.client.core.domain.Iso3Codes;
import com.coherentlogic.wb.client.core.domain.LendingTypeCodes;
import com.coherentlogic.wb.client.core.domain.RegionCodes;
import com.coherentlogic.wb.client.core.exceptions.InvalidFromToFormatException;
import com.coherentlogic.wb.client.core.exceptions.InvalidIsoCodeException;
import com.coherentlogic.wb.client.core.exceptions.InvalidMetatypesException;
import com.coherentlogic.wb.client.core.exceptions.InvalidParameterValueException;

/**
 * Class that allows the developer to construct and execute a query to the
 * World Bank web services.
 * <p>
 * Note that this class is <b>not</b> thread-safe and cannot be used as a member
 * -level property -- if this is required, use the
 * {@link com.coherentlogic.fred.client.core.builders#RequestBuilderFactory
 * QueryBuilderFactory} class.
 * <p>
 * In order to facilitate method-chaining each setter method returns a reference
 * to this object.
 * <p>
 * For examples, refer to the {@link
 * com.coherentlogic.wb.client.core.builders#QueryBuilderTest
 * RequestBuilderTest} class.
 *
 * @see http://data.worldbank.org
 * @see http://data.worldbank.org/developers/data-catalog-api
 * @see http://data.worldbank.org/querybuilder
 *
 * @author <a href="mailto:support@coherentlogic.com">Support</a>
 *
 * @todo Refactor this class as some of the logic needs to be shared between
 *  this class and the fred client.
 */
public class QueryBuilder extends AbstractQueryBuilder {

    /**
     * The year pattern -- ie. YYYY.
     */
    private static final String YEAR_PATTERN = "\\d{4}";

    /**
     * The from/to constant -- ie. YYYY:YYYY.
     */
    private static final String FROM_TO_PATTERN =
        YEAR_PATTERN + ":" + YEAR_PATTERN;

    /**
     * The from/to pattern is used to ensure that from/to strings have the
     * proper format.
     */
    private static final Pattern fromToPattern =
        Pattern.compile(FROM_TO_PATTERN);

    /**
     * @see {@link AbstractQueryBuilder}.
     */
    public QueryBuilder (
        RestTemplate restTemplate,
        String uri
    ) {
        super (restTemplate, uri);
    }

    /**
     * @see {@link AbstractQueryBuilder}.
     */
    public QueryBuilder (
        RestTemplate restTemplate,
        UriBuilder uriBuilder
    ) {
        super (restTemplate, uriBuilder);
    }

    /**
     * Extends the path to include countries -- ie:
     *
     * http://api.worldbank.org/countries
     */
    public QueryBuilder setCountries () {
        super.extendPathWith(COUNTRIES);

        return this;
    }

    /**
     * Extends the path to include source -- ie:
     *
     * http://api.worldbank.org/source
     */
    public QueryBuilder setSource () {
        super.extendPathWith(SOURCE);

        return this;
    }

    /**
     * Extends the path to include sources -- ie:
     *
     * http://api.worldbank.org/sources
     */
    public QueryBuilder setSources () {
        super.extendPathWith(SOURCES);

        return this;
    }

    /**
     * Extends the path to include incomeLevel -- ie:
     *
     * http://api.worldbank.org/incomelevel
     */
    public QueryBuilder setIncomeLevel () {
        super.extendPathWith(INCOME_LEVEL);

        return this;
    }

    /**
     * Extends the path to include incomeLevels -- ie:
     *
     * http://api.worldbank.org/incomelevels
     */
    public QueryBuilder setIncomeLevels () {
        super.extendPathWith(INCOME_LEVELS);

        return this;
    }

    /**
     * Extends the path to include indicators -- ie:
     *
     * http://api.worldbank.org/indicators
     */
    public QueryBuilder setIndicators () {
        super.extendPathWith(INDICATORS);

        return this;
    }

    /**
     * Extends the path to include indicator -- ie:
     *
     * http://api.worldbank.org/indicator
     */
    public QueryBuilder setIndicator () {
        super.extendPathWith(INDICATOR);

        return this;
    }

    /**
     * Extends the path to include lending types -- ie:
     *
     * http://api.worldbank.org/lendingTypes
     */
    public QueryBuilder setLendingTypes () {
        super.extendPathWith(LENDING_TYPES);

        return this;
    }

    /**
     * Extends the path to include topics -- ie:
     *
     * http://api.worldbank.org/topics
     */
    public QueryBuilder setTopics () {
        super.extendPathWith(TOPICS);

        return this;
    }

    /**
     * Extends the path to include topic -- ie:
     *
     * http://api.worldbank.org/topic
     */
    public QueryBuilder setTopic () {
        super.extendPathWith(TOPIC);

        return this;
    }

    /**
     * Extends the path to include data catalog -- ie:
     *
     * http://api.worldbank.org/datacatalog
     */
    public QueryBuilder setDataCatalog () {
        super.extendPathWith(DATA_CATALOG);

        return this;
    }

    /**
     * Extends the path to include the value of the path -- ie: when the current
     * URL is http://api.worldbank.org/, invoking this method with the value
     * "foo" results in the following:
     *
     * http://api.worldbank.org/foo
     */
    @Override
    public QueryBuilder extendPathWith (String path) {
        super.extendPathWith(path);

        return this;
    }

    /**
     * Method checks that the fromTo format is correct and, if it isn't, an
     * exception will be thrown.
     *
     * @param fromTo A string which should be in the format YYYY:YYYY.
     *
     * @throws InvalidFromToFormatException when the string is not valid.
     */
    private void reviewFromToFormat (String fromTo) {
        Matcher fromToMatcher = fromToPattern.matcher(fromTo);

        if (!fromToMatcher.matches())
            throw new InvalidFromToFormatException("The fromTo value is " +
                "not valid -- this value must be in the format YYYYY:YYYY" +
                "-- ie. '2001:2005'. (fromTo: " + fromTo + ").");
    }

    /**
     * Setter method for the date, which should be in the format YYYY:YYYY.
     *
     * @param fromTo The date range -- ie. 1998:2006.
     */
    public QueryBuilder setDate (String fromTo) {

        reviewFromToFormat (fromTo);

        addParameter(Constants.DATE, fromTo);

        return this;
    }

    /**
     * Setter method for the date.
     *
     * @param from For example, 1999.
     *
     * @param to For example, 2002.
     */
    public QueryBuilder setDate (String from, String to) {
        return setDate (from + ":" + to);
    }

    /**
     * Method throws an exception if the actual value is less than or equal to
     * the target value.
     */
    static void assertNotLessThanEqualTo (
        String param, int target, int actual) {
        if (actual <= target)
            throw new InvalidParameterValueException("The value of the " +
                "parameter named '" + param + "' is " + actual + " and this " +
                "is less than or equal to the expected target value of " +
                target);
    }

    /**
     * Setter method for the page parameter.
     *
     * @param page The page value.
     *
     * @todo The assumption at this point is that the page should be an int.
     */
    public QueryBuilder setPage (int page) {

        assertNotLessThanEqualTo ("page", 0, page);

        addParameter (PAGE, Integer.toString(page));

        return this;
    }

    /**
     * Setter method for the page value.
     *
     * @param perPage The perPage value.
     *
     * @todo Do we even need this?
     * @todo We need to check for values that are negative or zero.
     */
    public QueryBuilder setPerPage (int perPage) {

        assertNotLessThanEqualTo ("perPage", 0, perPage);

        addParameter (PER_PAGE, Integer.toString(perPage));

        return this;
    }

    /**
     * Setter method for the most recent values based on the number specified.
     *
     * @param mrv The most recent values.
     *
     * @see http://data.worldbank.org/node/11
     *
     * @todo Should this be an int?
     * @todo We need to check for zero and negative values.
     */
    public QueryBuilder setMRV (int mrv) {

        addParameter (MRV, Integer.toString(mrv));

        return this;
    }

    /**
     * Method sets the gapFill value to either 'yes' or 'no'.
     *
     * @param yesNo True equated to 'yes', 'no' otherwise.
     */
    public QueryBuilder setGapFill (boolean yesNo) {

        String value = yesNo == true ? YES : NO;

        setGapFill (value);

        return this;
    }

    /**
     * The yesNo value is either 'yes' or 'no'.
     */
    QueryBuilder setGapFill (String yesNo) {

        addParameter (GAP_FILL, yesNo);

        return this;
    }

    /**
     * Setter method for the frequency.
     */
    public QueryBuilder setFrequency (Frequency frequency) {

        if (frequency == null)
            throw new InvalidParameterValueException("The frequency is null.");

        addParameter (FREQUENCY, frequency.toString());

        return this;
    }

    /**
     * Setter method for the income level.
     */
    QueryBuilder setIncomeLevel (String incomeLevel) {

        addParameter (INCOME_LEVEL, incomeLevel);

        return this;
    }

    /**
     * Setter method for the lending type.
     */
    public QueryBuilder setLendingType (LendingTypeCodes lendingTypeCode) {

        if (lendingTypeCode == null)
            throw new InvalidParameterValueException(
                "The lendingTypeCode is null.");

        return setLendingType(lendingTypeCode.name());
    }

    /**
     * Setter method for the lending type.
     */
    QueryBuilder setLendingType (String lendingType) {

        addParameter (LENDING_TYPE, lendingType);

        return this;
    }

    /**
     * Setter method for the region.
     */
    public QueryBuilder setRegion (RegionCodes regionCode) {

        addParameter (REGION, regionCode.name());

        return this;
    }

    /**
     * Setter method for the income level.
     */
    public QueryBuilder setIncomeLevel (IncomeLevelCodes incomeLevelCode) {

        if (incomeLevelCode == null)
            throw new InvalidParameterValueException(
                "The incomeLevelCode is null.");

        return setIncomeLevel(incomeLevelCode.name());
    }

    /**
     * Method combines the array of values into a single string separated by
     * semicolons.
     */
    static String combine (String... values) {
        if (values == null || values.length == 0)
            throw new InvalidMetatypesException(
                "The values passed to the combine method are invalid -- " +
                "these values include: " +
                ToStringBuilder.reflectionToString(values));

        String result = "";

        int ctr = (values.length - 1);

        for (String value : values) {

            result += value;

            if (0 < ctr--)
                result += ";";
        }
        return result;
    }

    /**
     * Combines the isoCodes into an array of String values.
     *
     * @todo Combine this method with the one below into converter classes.
     */
    static String[] toStringArray (Iso2Codes... isoCodes) {

        if (isoCodes == null || isoCodes.length == 0)
            throw new InvalidIsoCodeException("The isoCodes parameter is " +
                "either null or empty (isoCodes: " + isoCodes + ").");

        String[] results =
            (String[]) ConvertUtils.convert(isoCodes, String[].class);

        return results;
    }

    /**
     * Combines the isoCodes into an array of String values.
     *
     * @todo Combine this method with the one above into converter classes.
     */
    static String[] toStringArray (Iso3Codes... isoCodes) {

        if (isoCodes == null || isoCodes.length == 0)
            throw new InvalidIsoCodeException("The isoCodes parameter is " +
                "either null or empty (isoCodes: " + isoCodes + ").");

        String[] results =
            (String[]) ConvertUtils.convert(isoCodes, String[].class);

        return results;
    }

    /**
     * Sets the isoCode in the URI.
     *
     * http://api.worldbank.org/countries/isoCode/
     *
     * @param isoCodes The isoCode.
     *
     * @see http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
     * @see http://en.wikipedia.org/wiki/ISO_3166-1_alpha-3
     * @see http://data.worldbank.org/node/18
     *
     * @todo Determine if we need to set the iso3 code.
     */
    QueryBuilder setIsoCodes (String... isoCodes) {

        String combinedIsoCodes = combine(isoCodes);

        super.extendPathWith(combinedIsoCodes);

        return this;
    }

    /**
     * Sets the Iso2Codes values in the URI.
     *
     * http://api.worldbank.org/countries/isoCode/
     *
     * @param isoCodes The isoCode.
     *
     * @see http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
     * @see http://en.wikipedia.org/wiki/ISO_3166-1_alpha-3
     * @see http://data.worldbank.org/node/18
     *
     * @todo Determine if we need to set the iso2 code.
     */
    public QueryBuilder setIsoCodes (Iso2Codes... isoCodes) {
        return setIsoCodes (toStringArray (isoCodes));
    }

    /**
     * Sets the Iso3Codes values in the URI.
     *
     * http://api.worldbank.org/countries/isoCode/
     *
     * @param isoCodes The isoCode.
     *
     * @see http://en.wikipedia.org/wiki/ISO_3166-1_alpha-2
     * @see http://en.wikipedia.org/wiki/ISO_3166-1_alpha-3
     * @see http://data.worldbank.org/node/18
     *
     * @todo Determine if we need to set the iso3 code.
     */
    public QueryBuilder setIsoCodes (Iso3Codes... isoCodes) {
        return setIsoCodes (toStringArray (isoCodes));
    }

    /**
     * Setter method for the metatypes parameter.
     *
     * @see http://data.worldbank.org/developers/data-catalog-api
     */
    public QueryBuilder setMetatypes (String... values) {

        String metatypes = combine (values);

        addParameter (METATYPES, metatypes);

        return this;
    }
}
